package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.Invocation;
import org.eclipse.jdt.internal.compiler.ast.ReferenceExpression;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BoundSet;
import org.eclipse.jdt.internal.compiler.lookup.InferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.InferenceContext18;
import org.eclipse.jdt.internal.compiler.lookup.InferenceFailureException;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.PolyParameterizedGenericMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.RawTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.Substitution;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;

public class ParameterizedGenericMethodBinding extends ParameterizedMethodBinding implements Substitution {
  public TypeBinding[] typeArguments;
  
  protected LookupEnvironment environment;
  
  public boolean inferredReturnType;
  
  public boolean wasInferred;
  
  public boolean isRaw;
  
  private MethodBinding tiebreakMethod;
  
  public boolean inferredWithUncheckedConversion;
  
  public static MethodBinding computeCompatibleMethod(MethodBinding originalMethod, TypeBinding[] arguments, Scope scope, InvocationSite invocationSite) {
    org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding methodSubstitute;
    TypeVariableBinding[] typeVariables = originalMethod.typeVariables;
    TypeBinding[] substitutes = invocationSite.genericTypeArguments();
    InferenceContext inferenceContext = null;
    TypeBinding[] uncheckedArguments = null;
    if (substitutes != null) {
      if (substitutes.length != typeVariables.length)
        return (MethodBinding)new ProblemMethodBinding(originalMethod, originalMethod.selector, substitutes, 11); 
      methodSubstitute = scope.environment().createParameterizedGenericMethod(originalMethod, substitutes);
    } else {
      TypeBinding[] parameters = originalMethod.parameters;
      CompilerOptions compilerOptions = scope.compilerOptions();
      if (compilerOptions.sourceLevel >= 3407872L)
        return computeCompatibleMethod18(originalMethod, arguments, scope, invocationSite); 
      inferenceContext = new InferenceContext(originalMethod);
      methodSubstitute = inferFromArgumentTypes(scope, originalMethod, arguments, parameters, inferenceContext);
      if (methodSubstitute == null)
        return null; 
      if (inferenceContext.hasUnresolvedTypeArgument()) {
        if (inferenceContext.isUnchecked) {
          int j = inferenceContext.substitutes.length;
          System.arraycopy(inferenceContext.substitutes, 0, uncheckedArguments = new TypeBinding[j], 0, j);
        } 
        if (methodSubstitute.returnType != TypeBinding.VOID) {
          ReferenceBinding referenceBinding;
          TypeBinding expectedType = invocationSite.invocationTargetType();
          if (expectedType != null) {
            inferenceContext.hasExplicitExpectedType = true;
          } else {
            referenceBinding = scope.getJavaLangObject();
          } 
          inferenceContext.expectedType = (TypeBinding)referenceBinding;
        } 
        methodSubstitute = methodSubstitute.inferFromExpectedType(scope, inferenceContext);
        if (methodSubstitute == null)
          return null; 
      } else if (compilerOptions.sourceLevel == 3342336L) {
        if (methodSubstitute.returnType != TypeBinding.VOID) {
          TypeBinding expectedType = invocationSite.invocationTargetType();
          if (expectedType != null && !originalMethod.returnType.mentionsAny(originalMethod.parameters, -1)) {
            TypeBinding uncaptured = methodSubstitute.returnType.uncapture(scope);
            if (!methodSubstitute.returnType.isCompatibleWith(expectedType) && 
              expectedType.isCompatibleWith(uncaptured)) {
              InferenceContext oldContext = inferenceContext;
              inferenceContext = new InferenceContext(originalMethod);
              originalMethod.returnType.collectSubstitutes(scope, expectedType, inferenceContext, 1);
              org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding substitute = inferFromArgumentTypes(scope, originalMethod, arguments, parameters, inferenceContext);
              if (substitute != null && substitute.returnType.isCompatibleWith(expectedType)) {
                if (scope.parameterCompatibilityLevel((MethodBinding)substitute, arguments, false) > -1) {
                  methodSubstitute = substitute;
                } else {
                  inferenceContext = oldContext;
                } 
              } else {
                inferenceContext = oldContext;
              } 
            } 
          } 
        } 
      } 
    } 
    Substitution substitution = null;
    if (inferenceContext != null) {
      LingeringTypeVariableEliminator lingeringTypeVariableEliminator = new LingeringTypeVariableEliminator(typeVariables, inferenceContext.substitutes, scope);
    } else {
      substitution = methodSubstitute;
    } 
    for (int i = 0, length = typeVariables.length; i < length; i++) {
      TypeBinding substituteForChecks;
      TypeVariableBinding typeVariable = typeVariables[i];
      TypeBinding substitute = methodSubstitute.typeArguments[i];
      if (substitute instanceof TypeVariableBinding) {
        substituteForChecks = substitute;
      } else {
        substituteForChecks = Scope.substitute((Substitution)new LingeringTypeVariableEliminator(typeVariables, null, scope), substitute);
      } 
      if (uncheckedArguments == null || uncheckedArguments[i] != null) {
        int argLength;
        TypeBinding[] augmentedArguments;
        switch (typeVariable.boundCheck(substitution, substituteForChecks, scope, null)) {
          case null:
            argLength = arguments.length;
            augmentedArguments = new TypeBinding[argLength + 2];
            System.arraycopy(arguments, 0, augmentedArguments, 0, argLength);
            augmentedArguments[argLength] = substitute;
            augmentedArguments[argLength + 1] = (TypeBinding)typeVariable;
            return (MethodBinding)new ProblemMethodBinding((MethodBinding)methodSubstitute, originalMethod.selector, augmentedArguments, 10);
          case UNCHECKED:
            methodSubstitute.tagBits |= 0x100L;
            break;
        } 
      } 
    } 
    return (MethodBinding)methodSubstitute;
  }
  
  public static MethodBinding computeCompatibleMethod18(MethodBinding originalMethod, TypeBinding[] arguments, Scope scope, InvocationSite invocationSite) {
    TypeVariableBinding[] typeVariables = originalMethod.typeVariables;
    if (invocationSite.checkingPotentialCompatibility())
      return (MethodBinding)scope.environment().createParameterizedGenericMethod(originalMethod, (TypeBinding[])typeVariables); 
    org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding methodSubstitute = null;
    InferenceContext18 infCtx18 = invocationSite.freshInferenceContext(scope);
    if (infCtx18 == null)
      return originalMethod; 
    TypeBinding[] parameters = originalMethod.parameters;
    CompilerOptions compilerOptions = scope.compilerOptions();
    boolean invocationTypeInferred = false;
    boolean requireBoxing = false;
    TypeBinding[] argumentsCopy = new TypeBinding[arguments.length];
    for (int i = 0, length = arguments.length, parametersLength = parameters.length; i < length; i++) {
      TypeBinding parameter = (i < parametersLength) ? parameters[i] : parameters[parametersLength - 1];
      TypeBinding argument = arguments[i];
      if (argument.isPrimitiveType() != parameter.isPrimitiveType()) {
        argumentsCopy[i] = scope.environment().computeBoxingType(argument);
        requireBoxing = true;
      } else {
        argumentsCopy[i] = argument;
      } 
    } 
    arguments = argumentsCopy;
    LookupEnvironment environment = scope.environment();
    InferenceContext18 previousContext = environment.currentInferenceContext;
    if (previousContext == null)
      environment.currentInferenceContext = infCtx18; 
    try {
      int j;
      BoundSet provisionalResult = null;
      BoundSet result = null;
      boolean isPolyExpression = (invocationSite instanceof Expression && ((Expression)invocationSite).isPolyExpression(originalMethod));
      boolean isDiamond = (isPolyExpression && originalMethod.isConstructor());
      if (arguments.length == parameters.length) {
        infCtx18.inferenceKind = requireBoxing ? 2 : 1;
        infCtx18.inferInvocationApplicability(originalMethod, arguments, isDiamond);
        result = infCtx18.solve(true);
      } 
      if (result == null && originalMethod.isVarargs()) {
        infCtx18 = invocationSite.freshInferenceContext(scope);
        infCtx18.inferenceKind = 3;
        infCtx18.inferInvocationApplicability(originalMethod, arguments, isDiamond);
        result = infCtx18.solve(true);
      } 
      if (result == null)
        return null; 
      if (infCtx18.isResolved(result)) {
        infCtx18.stepCompleted = 1;
      } else {
        return null;
      } 
      TypeBinding expectedType = invocationSite.invocationTargetType();
      boolean hasReturnProblem = false;
      if (expectedType != null || !invocationSite.getExpressionContext().definesTargetType() || !isPolyExpression) {
        provisionalResult = result;
        result = infCtx18.inferInvocationType(expectedType, invocationSite, originalMethod);
        invocationTypeInferred = true;
        j = hasReturnProblem | ((result == null) ? 1 : 0);
        if (j != 0)
          result = provisionalResult; 
      } 
      if (result != null) {
        TypeBinding[] solutions = infCtx18.getSolutions(typeVariables, invocationSite, result);
        if (solutions != null) {
          PolyParameterizedGenericMethodBinding polyParameterizedGenericMethodBinding;
          methodSubstitute = scope.environment().createParameterizedGenericMethod(originalMethod, solutions, infCtx18.usesUncheckedConversion, j);
          if (invocationSite instanceof Invocation)
            infCtx18.forwardResults(result, (Invocation)invocationSite, methodSubstitute, expectedType); 
          try {
            if (j != 0) {
              MethodBinding problemMethod = infCtx18.getReturnProblemMethodIfNeeded(expectedType, (MethodBinding)methodSubstitute);
              if (problemMethod instanceof ProblemMethodBinding)
                return problemMethod; 
            } 
          } finally {
            if (invocationSite instanceof Invocation) {
              ((Invocation)invocationSite).registerInferenceContext((org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding)polyParameterizedGenericMethodBinding, infCtx18);
            } else if (invocationSite instanceof ReferenceExpression) {
              ((ReferenceExpression)invocationSite).registerInferenceContext((org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding)polyParameterizedGenericMethodBinding, infCtx18);
            } 
          } 
          if (invocationSite instanceof Invocation) {
            ((Invocation)invocationSite).registerInferenceContext((org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding)polyParameterizedGenericMethodBinding, infCtx18);
          } else if (invocationSite instanceof ReferenceExpression) {
            ((ReferenceExpression)invocationSite).registerInferenceContext((org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding)polyParameterizedGenericMethodBinding, infCtx18);
          } 
          return (MethodBinding)polyParameterizedGenericMethodBinding;
        } 
      } 
      return null;
    } catch (InferenceFailureException e) {
      scope.problemReporter().genericInferenceError(e.getMessage(), invocationSite);
      return null;
    } finally {
      environment.currentInferenceContext = previousContext;
    } 
  }
  
  MethodBinding boundCheck18(Scope scope, TypeBinding[] arguments, InvocationSite site) {
    Substitution substitution = this;
    org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding methodSubstitute = this;
    TypeVariableBinding[] originalTypeVariables = this.originalMethod.typeVariables;
    for (int i = 0, length = originalTypeVariables.length; i < length; i++) {
      int argLength;
      TypeBinding[] augmentedArguments;
      TypeVariableBinding typeVariable = originalTypeVariables[i];
      TypeBinding substitute = methodSubstitute.typeArguments[i];
      ASTNode location = (site instanceof ASTNode) ? (ASTNode)site : null;
      switch (typeVariable.boundCheck(substitution, substitute, scope, location)) {
        case null:
          argLength = arguments.length;
          augmentedArguments = new TypeBinding[argLength + 2];
          System.arraycopy(arguments, 0, augmentedArguments, 0, argLength);
          augmentedArguments[argLength] = substitute;
          augmentedArguments[argLength + 1] = (TypeBinding)typeVariable;
          return (MethodBinding)new ProblemMethodBinding((MethodBinding)methodSubstitute, this.originalMethod.selector, augmentedArguments, 10);
        case UNCHECKED:
          methodSubstitute.tagBits |= 0x100L;
          break;
      } 
    } 
    return null;
  }
  
  private static org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding inferFromArgumentTypes(Scope scope, MethodBinding originalMethod, TypeBinding[] arguments, TypeBinding[] parameters, InferenceContext inferenceContext) {
    // Byte code:
    //   0: aload_1
    //   1: invokevirtual isVarargs : ()Z
    //   4: ifeq -> 202
    //   7: aload_3
    //   8: arraylength
    //   9: istore #5
    //   11: iload #5
    //   13: iconst_1
    //   14: isub
    //   15: istore #6
    //   17: aload_2
    //   18: arraylength
    //   19: istore #7
    //   21: iconst_0
    //   22: istore #8
    //   24: goto -> 56
    //   27: aload_3
    //   28: iload #8
    //   30: aaload
    //   31: aload_0
    //   32: aload_2
    //   33: iload #8
    //   35: aaload
    //   36: aload #4
    //   38: iconst_1
    //   39: invokevirtual collectSubstitutes : (Lorg/eclipse/jdt/internal/compiler/lookup/Scope;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/InferenceContext;I)V
    //   42: aload #4
    //   44: getfield status : I
    //   47: iconst_1
    //   48: if_icmpne -> 53
    //   51: aconst_null
    //   52: areturn
    //   53: iinc #8, 1
    //   56: iload #8
    //   58: iload #6
    //   60: if_icmplt -> 27
    //   63: iload #6
    //   65: iload #7
    //   67: if_icmpge -> 248
    //   70: aload_3
    //   71: iload #6
    //   73: aaload
    //   74: astore #8
    //   76: aload_2
    //   77: iload #6
    //   79: aaload
    //   80: astore #9
    //   82: iload #5
    //   84: iload #7
    //   86: if_icmpne -> 148
    //   89: aload #9
    //   91: getstatic org/eclipse/jdt/internal/compiler/lookup/TypeBinding.NULL : Lorg/eclipse/jdt/internal/compiler/lookup/NullTypeBinding;
    //   94: if_acmpne -> 100
    //   97: goto -> 158
    //   100: aload #9
    //   102: invokevirtual dimensions : ()I
    //   105: tableswitch default -> 145, 0 -> 128, 1 -> 131
    //   128: goto -> 148
    //   131: aload #9
    //   133: invokevirtual leafComponentType : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   136: invokevirtual isBaseType : ()Z
    //   139: ifne -> 148
    //   142: goto -> 158
    //   145: goto -> 158
    //   148: aload #8
    //   150: checkcast org/eclipse/jdt/internal/compiler/lookup/ArrayBinding
    //   153: invokevirtual elementsType : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   156: astore #8
    //   158: iload #6
    //   160: istore #10
    //   162: goto -> 192
    //   165: aload #8
    //   167: aload_0
    //   168: aload_2
    //   169: iload #10
    //   171: aaload
    //   172: aload #4
    //   174: iconst_1
    //   175: invokevirtual collectSubstitutes : (Lorg/eclipse/jdt/internal/compiler/lookup/Scope;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/InferenceContext;I)V
    //   178: aload #4
    //   180: getfield status : I
    //   183: iconst_1
    //   184: if_icmpne -> 189
    //   187: aconst_null
    //   188: areturn
    //   189: iinc #10, 1
    //   192: iload #10
    //   194: iload #7
    //   196: if_icmplt -> 165
    //   199: goto -> 248
    //   202: aload_3
    //   203: arraylength
    //   204: istore #5
    //   206: iconst_0
    //   207: istore #6
    //   209: goto -> 241
    //   212: aload_3
    //   213: iload #6
    //   215: aaload
    //   216: aload_0
    //   217: aload_2
    //   218: iload #6
    //   220: aaload
    //   221: aload #4
    //   223: iconst_1
    //   224: invokevirtual collectSubstitutes : (Lorg/eclipse/jdt/internal/compiler/lookup/Scope;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/InferenceContext;I)V
    //   227: aload #4
    //   229: getfield status : I
    //   232: iconst_1
    //   233: if_icmpne -> 238
    //   236: aconst_null
    //   237: areturn
    //   238: iinc #6, 1
    //   241: iload #6
    //   243: iload #5
    //   245: if_icmplt -> 212
    //   248: aload_1
    //   249: getfield typeVariables : [Lorg/eclipse/jdt/internal/compiler/lookup/TypeVariableBinding;
    //   252: astore #5
    //   254: aload_0
    //   255: aload #5
    //   257: aload #4
    //   259: iconst_0
    //   260: invokestatic resolveSubstituteConstraints : (Lorg/eclipse/jdt/internal/compiler/lookup/Scope;[Lorg/eclipse/jdt/internal/compiler/lookup/TypeVariableBinding;Lorg/eclipse/jdt/internal/compiler/lookup/InferenceContext;Z)Z
    //   263: ifne -> 268
    //   266: aconst_null
    //   267: areturn
    //   268: aload #4
    //   270: getfield substitutes : [Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   273: astore #6
    //   275: aload #6
    //   277: astore #7
    //   279: iconst_0
    //   280: istore #8
    //   282: aload #5
    //   284: arraylength
    //   285: istore #9
    //   287: goto -> 355
    //   290: aload #6
    //   292: iload #8
    //   294: aaload
    //   295: ifnonnull -> 335
    //   298: aload #7
    //   300: aload #6
    //   302: if_acmpne -> 322
    //   305: aload #6
    //   307: iconst_0
    //   308: iload #9
    //   310: anewarray org/eclipse/jdt/internal/compiler/lookup/TypeBinding
    //   313: dup
    //   314: astore #7
    //   316: iconst_0
    //   317: iload #8
    //   319: invokestatic arraycopy : (Ljava/lang/Object;ILjava/lang/Object;II)V
    //   322: aload #7
    //   324: iload #8
    //   326: aload #5
    //   328: iload #8
    //   330: aaload
    //   331: aastore
    //   332: goto -> 352
    //   335: aload #7
    //   337: aload #6
    //   339: if_acmpeq -> 352
    //   342: aload #7
    //   344: iload #8
    //   346: aload #6
    //   348: iload #8
    //   350: aaload
    //   351: aastore
    //   352: iinc #8, 1
    //   355: iload #8
    //   357: iload #9
    //   359: if_icmplt -> 290
    //   362: aload_0
    //   363: invokevirtual environment : ()Lorg/eclipse/jdt/internal/compiler/lookup/LookupEnvironment;
    //   366: aload_1
    //   367: aload #7
    //   369: invokevirtual createParameterizedGenericMethod : (Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;[Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Lorg/eclipse/jdt/internal/compiler/lookup/ParameterizedGenericMethodBinding;
    //   372: astore #8
    //   374: aload #8
    //   376: areturn
    // Line number table:
    //   Java source line number -> byte code offset
    //   #343	-> 0
    //   #344	-> 7
    //   #345	-> 11
    //   #346	-> 17
    //   #348	-> 21
    //   #349	-> 27
    //   #350	-> 42
    //   #348	-> 53
    //   #353	-> 63
    //   #354	-> 70
    //   #355	-> 76
    //   #357	-> 82
    //   #358	-> 89
    //   #359	-> 100
    //   #361	-> 128
    //   #363	-> 131
    //   #366	-> 145
    //   #370	-> 148
    //   #372	-> 158
    //   #373	-> 165
    //   #374	-> 178
    //   #372	-> 189
    //   #377	-> 199
    //   #378	-> 202
    //   #379	-> 206
    //   #380	-> 212
    //   #381	-> 227
    //   #379	-> 238
    //   #384	-> 248
    //   #385	-> 254
    //   #386	-> 266
    //   #389	-> 268
    //   #390	-> 275
    //   #391	-> 279
    //   #392	-> 290
    //   #393	-> 298
    //   #394	-> 305
    //   #396	-> 322
    //   #397	-> 332
    //   #398	-> 342
    //   #391	-> 352
    //   #401	-> 362
    //   #402	-> 374
    // Local variable table:
    //   start	length	slot	name	descriptor
    //   0	377	0	scope	Lorg/eclipse/jdt/internal/compiler/lookup/Scope;
    //   0	377	1	originalMethod	Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   0	377	2	arguments	[Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   0	377	3	parameters	[Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   0	377	4	inferenceContext	Lorg/eclipse/jdt/internal/compiler/lookup/InferenceContext;
    //   11	188	5	paramLength	I
    //   17	182	6	minArgLength	I
    //   21	178	7	argLength	I
    //   24	39	8	i	I
    //   76	123	8	varargType	Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   82	117	9	lastArgument	Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   162	37	10	i	I
    //   206	42	5	paramLength	I
    //   209	39	6	i	I
    //   254	123	5	originalVariables	[Lorg/eclipse/jdt/internal/compiler/lookup/TypeVariableBinding;
    //   275	102	6	inferredSustitutes	[Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   279	98	7	actualSubstitutes	[Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   282	80	8	i	I
    //   287	75	9	varLength	I
    //   374	3	8	paramMethod	Lorg/eclipse/jdt/internal/compiler/lookup/ParameterizedGenericMethodBinding;
  }
  
  private static boolean resolveSubstituteConstraints(Scope scope, TypeVariableBinding[] typeVariables, InferenceContext inferenceContext, boolean considerEXTENDSConstraints) {
    TypeBinding[] substitutes = inferenceContext.substitutes;
    int varLength = typeVariables.length;
    int i;
    for (i = 0; i < varLength; i++) {
      TypeVariableBinding current = typeVariables[i];
      TypeBinding substitute = substitutes[i];
      if (substitute == null) {
        TypeBinding[] equalSubstitutes = inferenceContext.getSubstitutes(current, 0);
        if (equalSubstitutes != null)
          for (int j = 0, equalLength = equalSubstitutes.length; j < equalLength; ) {
            TypeBinding equalSubstitute = equalSubstitutes[j];
            if (equalSubstitute == null) {
              j++;
              continue;
            } 
            if (TypeBinding.equalsEquals(equalSubstitute, (TypeBinding)current)) {
              int k = j + 1;
              while (true) {
                if (k >= equalLength) {
                  substitutes[i] = (TypeBinding)current;
                  break;
                } 
                equalSubstitute = equalSubstitutes[k];
                if (TypeBinding.notEquals(equalSubstitute, (TypeBinding)current) && equalSubstitute != null) {
                  substitutes[i] = equalSubstitute;
                  break;
                } 
                k++;
              } 
              break;
            } 
            substitutes[i] = equalSubstitute;
            break;
          }  
      } 
    } 
    if (inferenceContext.hasUnresolvedTypeArgument())
      for (i = 0; i < varLength; i++) {
        TypeVariableBinding current = typeVariables[i];
        TypeBinding substitute = substitutes[i];
        if (substitute == null) {
          TypeBinding[] bounds = inferenceContext.getSubstitutes(current, 2);
          if (bounds != null) {
            TypeBinding mostSpecificSubstitute = scope.lowerUpperBound(bounds);
            if (mostSpecificSubstitute == null)
              return false; 
            if (mostSpecificSubstitute != TypeBinding.VOID)
              substitutes[i] = mostSpecificSubstitute; 
          } 
        } 
      }  
    if (considerEXTENDSConstraints && inferenceContext.hasUnresolvedTypeArgument())
      for (i = 0; i < varLength; i++) {
        TypeVariableBinding current = typeVariables[i];
        TypeBinding substitute = substitutes[i];
        if (substitute == null) {
          TypeBinding[] bounds = inferenceContext.getSubstitutes(current, 1);
          if (bounds != null) {
            WildcardBinding wildcardBinding;
            TypeBinding[] glb = Scope.greaterLowerBound(bounds, scope, scope.environment());
            TypeBinding mostSpecificSubstitute = null;
            if (glb != null)
              if (glb.length == 1) {
                mostSpecificSubstitute = glb[0];
              } else {
                TypeBinding[] otherBounds = new TypeBinding[glb.length - 1];
                System.arraycopy(glb, 1, otherBounds, 0, glb.length - 1);
                wildcardBinding = scope.environment().createWildcard(null, 0, glb[0], otherBounds, 1);
              }  
            if (wildcardBinding != null)
              substitutes[i] = (TypeBinding)wildcardBinding; 
          } 
        } 
      }  
    return true;
  }
  
  public ParameterizedGenericMethodBinding(MethodBinding originalMethod, RawTypeBinding rawType, LookupEnvironment environment) {
    TypeVariableBinding[] originalVariables = originalMethod.typeVariables;
    int length = originalVariables.length;
    TypeBinding[] rawArguments = new TypeBinding[length];
    for (int i = 0; i < length; i++)
      rawArguments[i] = environment.convertToRawType(originalVariables[i].erasure(), false); 
    this.isRaw = true;
    this.tagBits = originalMethod.tagBits;
    this.environment = environment;
    this.modifiers = originalMethod.modifiers;
    this.selector = originalMethod.selector;
    this.declaringClass = (rawType == null) ? originalMethod.declaringClass : (ReferenceBinding)rawType;
    this.typeVariables = Binding.NO_TYPE_VARIABLES;
    this.typeArguments = rawArguments;
    this.originalMethod = originalMethod;
    boolean ignoreRawTypeSubstitution = !(rawType != null && !originalMethod.isStatic());
    this.parameters = Scope.substitute(this, ignoreRawTypeSubstitution ? 
        originalMethod.parameters : 
        Scope.substitute((Substitution)rawType, originalMethod.parameters));
    this.thrownExceptions = Scope.substitute(this, ignoreRawTypeSubstitution ? 
        originalMethod.thrownExceptions : 
        Scope.substitute((Substitution)rawType, originalMethod.thrownExceptions));
    if (this.thrownExceptions == null)
      this.thrownExceptions = Binding.NO_EXCEPTIONS; 
    this.returnType = Scope.substitute(this, ignoreRawTypeSubstitution ? 
        originalMethod.returnType : 
        Scope.substitute((Substitution)rawType, originalMethod.returnType));
    this.wasInferred = false;
    this.parameterNonNullness = originalMethod.parameterNonNullness;
    this.defaultNullness = originalMethod.defaultNullness;
  }
  
  public ParameterizedGenericMethodBinding(MethodBinding originalMethod, TypeBinding[] typeArguments, LookupEnvironment environment, boolean inferredWithUncheckConversion, boolean hasReturnProblem) {
    this.environment = environment;
    this.inferredWithUncheckedConversion = inferredWithUncheckConversion;
    this.modifiers = originalMethod.modifiers;
    this.selector = originalMethod.selector;
    this.declaringClass = originalMethod.declaringClass;
    if (inferredWithUncheckConversion && originalMethod.isConstructor() && this.declaringClass.isParameterizedType())
      this.declaringClass = (ReferenceBinding)environment.convertToRawType(this.declaringClass.erasure(), false); 
    this.typeVariables = Binding.NO_TYPE_VARIABLES;
    this.typeArguments = typeArguments;
    this.isRaw = false;
    this.tagBits = originalMethod.tagBits;
    this.originalMethod = originalMethod;
    this.parameters = Scope.substitute(this, originalMethod.parameters);
    if (inferredWithUncheckConversion) {
      this.returnType = getErasure18_5_2(originalMethod.returnType, environment, hasReturnProblem);
      this.thrownExceptions = new ReferenceBinding[originalMethod.thrownExceptions.length];
      for (int j = 0; j < originalMethod.thrownExceptions.length; j++)
        this.thrownExceptions[j] = (ReferenceBinding)getErasure18_5_2((TypeBinding)originalMethod.thrownExceptions[j], environment, false); 
    } else {
      this.returnType = Scope.substitute(this, originalMethod.returnType);
      this.thrownExceptions = Scope.substitute(this, originalMethod.thrownExceptions);
    } 
    if (this.thrownExceptions == null)
      this.thrownExceptions = Binding.NO_EXCEPTIONS; 
    if ((this.tagBits & 0x80L) == 0L)
      if ((this.returnType.tagBits & 0x80L) != 0L) {
        this.tagBits |= 0x80L;
      } else {
        int j;
        int max;
        for (j = 0, max = this.parameters.length; j < max; j++) {
          if (((this.parameters[j]).tagBits & 0x80L) != 0L) {
            this.tagBits |= 0x80L;
            break label50;
          } 
        } 
        label50: for (j = 0, max = this.thrownExceptions.length; j < max; j++) {
          if (((this.thrownExceptions[j]).tagBits & 0x80L) != 0L) {
            this.tagBits |= 0x80L;
            break;
          } 
        } 
      }  
    this.wasInferred = true;
    this.parameterNonNullness = originalMethod.parameterNonNullness;
    this.defaultNullness = originalMethod.defaultNullness;
    int len = this.parameters.length;
    for (int i = 0; i < len; i++) {
      if (this.parameters[i] == TypeBinding.NULL) {
        long nullBits = (originalMethod.parameters[i]).tagBits & 0x180000000000000L;
        if (nullBits == 72057594037927936L) {
          if (this.parameterNonNullness == null)
            this.parameterNonNullness = new Boolean[len]; 
          this.parameterNonNullness[i] = Boolean.TRUE;
        } 
      } 
    } 
  }
  
  TypeBinding getErasure18_5_2(TypeBinding type, LookupEnvironment env, boolean substitute) {
    if (substitute)
      type = Scope.substitute(this, type); 
    return env.convertToRawType(type.erasure(), true);
  }
  
  public char[] computeUniqueKey(boolean isLeaf) {
    StringBuffer buffer = new StringBuffer();
    buffer.append(this.originalMethod.computeUniqueKey(false));
    buffer.append('%');
    buffer.append('<');
    if (!this.isRaw) {
      int length = this.typeArguments.length;
      for (int i = 0; i < length; i++) {
        TypeBinding typeArgument = this.typeArguments[i];
        buffer.append(typeArgument.computeUniqueKey(false));
      } 
    } 
    buffer.append('>');
    int resultLength = buffer.length();
    char[] result = new char[resultLength];
    buffer.getChars(0, resultLength, result, 0);
    return result;
  }
  
  public LookupEnvironment environment() {
    return this.environment;
  }
  
  public boolean hasSubstitutedParameters() {
    if (this.wasInferred)
      return this.originalMethod.hasSubstitutedParameters(); 
    return super.hasSubstitutedParameters();
  }
  
  public boolean hasSubstitutedReturnType() {
    if (this.inferredReturnType)
      return this.originalMethod.hasSubstitutedReturnType(); 
    return super.hasSubstitutedReturnType();
  }
  
  private org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding inferFromExpectedType(Scope scope, InferenceContext inferenceContext) {
    TypeVariableBinding[] originalVariables = this.originalMethod.typeVariables;
    int varLength = originalVariables.length;
    if (inferenceContext.expectedType != null) {
      this.returnType.collectSubstitutes(scope, inferenceContext.expectedType, inferenceContext, 2);
      if (inferenceContext.status == 1)
        return null; 
    } 
    int i;
    for (i = 0; i < varLength; i++) {
      TypeVariableBinding originalVariable = originalVariables[i];
      TypeBinding argument = this.typeArguments[i];
      boolean argAlreadyInferred = TypeBinding.notEquals(argument, (TypeBinding)originalVariable);
      if (TypeBinding.equalsEquals(originalVariable.firstBound, (TypeBinding)originalVariable.superclass)) {
        TypeBinding substitutedBound = Scope.substitute(this, (TypeBinding)originalVariable.superclass);
        argument.collectSubstitutes(scope, substitutedBound, inferenceContext, 2);
        if (inferenceContext.status == 1)
          return null; 
        if (argAlreadyInferred) {
          substitutedBound.collectSubstitutes(scope, argument, inferenceContext, 1);
          if (inferenceContext.status == 1)
            return null; 
        } 
      } 
      for (int j = 0, max = originalVariable.superInterfaces.length; j < max; j++) {
        TypeBinding substitutedBound = Scope.substitute(this, (TypeBinding)originalVariable.superInterfaces[j]);
        argument.collectSubstitutes(scope, substitutedBound, inferenceContext, 2);
        if (inferenceContext.status == 1)
          return null; 
        if (argAlreadyInferred) {
          substitutedBound.collectSubstitutes(scope, argument, inferenceContext, 1);
          if (inferenceContext.status == 1)
            return null; 
        } 
      } 
    } 
    if (!resolveSubstituteConstraints(scope, originalVariables, inferenceContext, true))
      return null; 
    for (i = 0; i < varLength; i++) {
      TypeBinding substitute = inferenceContext.substitutes[i];
      if (substitute != null) {
        this.typeArguments[i] = substitute;
      } else {
        inferenceContext.substitutes[i] = originalVariables[i].upperBound();
        this.typeArguments[i] = originalVariables[i].upperBound();
      } 
    } 
    this.typeArguments = Scope.substitute(this, this.typeArguments);
    TypeBinding oldReturnType = this.returnType;
    this.returnType = Scope.substitute(this, this.returnType);
    this.inferredReturnType = (inferenceContext.hasExplicitExpectedType && TypeBinding.notEquals(this.returnType, oldReturnType));
    this.parameters = Scope.substitute(this, this.parameters);
    this.thrownExceptions = Scope.substitute(this, this.thrownExceptions);
    if (this.thrownExceptions == null)
      this.thrownExceptions = Binding.NO_EXCEPTIONS; 
    if ((this.tagBits & 0x80L) == 0L)
      if ((this.returnType.tagBits & 0x80L) != 0L) {
        this.tagBits |= 0x80L;
      } else {
        int j;
        int max;
        for (j = 0, max = this.parameters.length; j < max; j++) {
          if (((this.parameters[j]).tagBits & 0x80L) != 0L) {
            this.tagBits |= 0x80L;
            break label71;
          } 
        } 
        label71: for (j = 0, max = this.thrownExceptions.length; j < max; j++) {
          if (((this.thrownExceptions[j]).tagBits & 0x80L) != 0L) {
            this.tagBits |= 0x80L;
            break;
          } 
        } 
      }  
    return this;
  }
  
  public boolean isParameterizedGeneric() {
    return true;
  }
  
  public boolean isRawSubstitution() {
    return this.isRaw;
  }
  
  public TypeBinding substitute(TypeVariableBinding originalVariable) {
    TypeVariableBinding[] variables = this.originalMethod.typeVariables;
    int length = variables.length;
    if (originalVariable.rank < length && TypeBinding.equalsEquals((TypeBinding)variables[originalVariable.rank], (TypeBinding)originalVariable)) {
      TypeBinding substitute = this.typeArguments[originalVariable.rank];
      return originalVariable.combineTypeAnnotations(substitute);
    } 
    return (TypeBinding)originalVariable;
  }
  
  public MethodBinding tiebreakMethod() {
    if (this.tiebreakMethod == null)
      this.tiebreakMethod = this.originalMethod.asRawMethod(this.environment); 
    return this.tiebreakMethod;
  }
  
  public MethodBinding genericMethod() {
    if (this.isRaw)
      return (MethodBinding)this; 
    return this.originalMethod;
  }
}
